"""
          Yasuharu Okamoto, 2021.12
"""
import numpy as np
import scipy.stats as ss
import scipy.integrate as si
from RNGen import *           #   Random number generator

CumPhi = ss.norm.cdf
phi = ss.norm.pdf

def f_phi(z):
    return phi(z)


class k_Pc:
    def __init__(self, d):
        self.d = d

    def f(self,u):
        """
               Frijters, Kooistra & Vereijken.(1980).
               Perception & Psychophysics, 1980, 27, 176-178.
        """
        v = (CumPhi(-u*(3**0.5) + self.d * ((2/3)**0.5))
             + CumPhi(-u*(3**0.5) - self.d * ((2/3)**0.5))) \
            * phi(u)
        return v

def f_Pc(d):
    int_k_Pc = k_Pc(d)
    v = si.quad(int_k_Pc.f, 0.0, +np.inf)
    return v[0] * 2.0


def calcPDprimes():
    """
            Dictionary:  key = d'
                         value = probability of the correct response
    """
    pdprimes = {}
    dprimes = np.linspace(0.0, 10.0, 1001)
    for i, d in enumerate(dprimes):
        if i % 100 == 0:
            print(i, end = '\r')
        p = f_Pc(d)
        pdprimes[f'{d:.2f}'] = p
    return pdprimes


def scale_sgm(acpt_r, ck):
    """
            Check the size of acceptance proportion
    """
    v = 1.0
    if (acpt_r > 0.5):
        ck += 1
        v = 1.0 + 9.0 * (acpt_r - 0.5) / 0.5
    elif (acpt_r < 0.3):
        ck += 1
        v = 1.0 / (1.0 + 9.0 * (0.3 - acpt_r) / 0.3)
    return v, ck

def my_log(v):
    if v <= 0.0:
        return -750.0
    else:
        return math.log(v)

class MCMC:
    """
                  Metropolis-within-Gibbs algorithm
                  cf. Robert,C.P. & Casella, G. (2010)
                  Monte Carlo Statistical Methods. p.393
                  岡本安晴 (2014)．心理学データ分析と測定．pp.229-232.
    """
    def __init__(self, Rs, Ags, TblPDprime, init_a, init_b,
                 sgm_a, sgm_b, n_samples, chain_id):
        self.Rs = Rs
        self.Ags = Ags
        self.TblPDprime = TblPDprime
        self.n_samples = n_samples
        self.chain_id = chain_id
        self.a_coeff = np.empty(self.n_samples + 1)
        self.b_coeff = np.empty(self.n_samples + 1)
        self.a_coeff[0] = init_a
        self.b_coeff[0] = init_b
        self.sgm_a = sgm_a
        self.sgm_b = sgm_b
        self.naccpt_a = 0
        self.naccpt_b = 0

    def LL(self, a, b):
        """
                Log likelihood of the parameters a and b
        """
        v = my_log(phi(a, 0.0, 10.0)) + my_log(phi(b, 0.0, 10.0))
        for i in range(len(self.Rs)):
            dprime = np.exp(a + b * self.Ags[i])
            if dprime > 10.0:
                dprime = 10.0
            p = self.TblPDprime[f'{dprime:.2f}']
            v += self.Rs[i] * my_log(p) +\
                 (1 - self.Rs[i]) * my_log(1- p)
               
        return v
    

    def do_mcmc(self):
        """
                Doing MCMC sampling
        """
        import tkinter as tk    #   For message window
        
        if self.chain_id < 0:
            msg0 = 'Adjusting sgm...step-{}'.format(-self.chain_id)
        else:
            msg0 = 'Proc-{}'.format(self.chain_id)
        
        root = tk.Tk()
        root.title(msg0)
        cnvs = tk.Canvas(width = 700, height = 300, background = '#00ffff')
        cnvs.pack()
        cnvs.create_text(350, 150, text = '', font = ('', 90), fill = 'blue',
                         tags = 'MyText')
        
        rn = RNp2to191()
        #
        #  Get independent chains
        #  cf. http://y-okamoto-psy1949.la.coocan.jp/Python/sampleprgs/rngeneration/en/
        #
        for i in range(self.chain_id): 
            rn.jump(100)
        
        prev_LL = self.LL(self.a_coeff[0], self.b_coeff[0])   #   The initial values
        den = prev_LL                                         #   The initial value
        for t in range(self.n_samples):
            if (t % 100) == 0:
                cnvs.itemconfig('MyText', text = f'{t}/{self.n_samples}')
                cnvs.update()
            #
            #   The next value for the parameter a
            #
            y = rn.normalMS(self.a_coeff[t], self.sgm_a)
            num = self.LL(y, self.b_coeff[t])
            den = prev_LL
            accpt = math.exp(num - den)
            if rn.uni() < accpt:
                self.a_coeff[t + 1] = y
                self.naccpt_a += 1
                prev_LL = num
            else:
                self.a_coeff[t + 1] = self.a_coeff[t]
                prev_LL = den
            #
            #   The next value for the parameter b
            #
            y = rn.normalMS(self.b_coeff[t], self.sgm_b)
            num = self.LL(self.a_coeff[t+1], y)
            den = prev_LL
            accpt = math.exp(num - den)
            if rn.uni() < accpt:
                self.b_coeff[t + 1] = y
                self.naccpt_b += 1
                prev_LL = num
            else:
                self.b_coeff[t + 1] = self.b_coeff[t]
                prev_LL = den


        root.destroy()

        return self.naccpt_a, self.naccpt_b, self.n_samples, \
               self.a_coeff[1:], self.b_coeff[1:]


def MCMCs(params):
    """
            The function for muliprocessing 
    """
    my_mcmc = MCMC(params['Rs'], params['Ags'], params['TblPDprime'],
                   params['init_a'],
                   params['init_b'], params['sgm_a'], params['sgm_b'],
                   params['n_samples'], params['chain_id'])
    r = my_mcmc.do_mcmc()
    return r



def calcRhat(psi):
    """
             Gelman,A., Carlin,J.B.,Stern,H.S., Dunson,D.B.,
             Vehtari,A, & Rubin,D.B. (2014)
             Bayesian Data Analysis, third edition. pp.284-285.
    """
    psi = np.array(psi)
    psidj = psi.mean(axis = 1)
    psidd = psi.mean()
    n = len(psi[0])
    m = len(psi)
    B = (((psidj - psidd) ** 2).sum()) * n / (m - 1)
    sj2 = []
    for j in range(m):
        sj2.append(((psi[j] - psidj[j]) ** 2).sum() / (n - 1))
    sj2 = np.array(sj2)
    W = (sj2.sum()) / m
    var_hat = W * (n - 1) / n + B / n
    Rhat = (var_hat / W) ** 0.5
    return Rhat        

    
def EstMode(samples, a = 0.05, n_points = 10000):
    """
        Calculatte a MAP estimate from a KDE graph on [Lp, Up]
        Lp and Up are 100*a/2 and 100(1-a/2) percentile points of samples
    """
    import numpy as np
    import scipy.stats as ss
    Lp, Up = np.percentile(samples, [100 * a/2, 100 * (1 - a/2)])  # import numpy as np
    coord = np.linspace(Lp, Up, n_points)  
    est_pdf = ss.gaussian_kde(samples).pdf(coord)  # import scipy.stats as ss
    mode_idx = np.argmax(est_pdf)  
    Mode_Est = coord[mode_idx]        
    return Mode_Est
